import discord
import os, random, io
from datetime import datetime
from fuzzywuzzy import fuzz, process
from minio import Minio
from minio.commonconfig import Tags
from minio.error import S3Error
from catbox import Uploader
from hashlib import md5

# memes we have most recently gotten
class RecentMemeQueue:
    #last_memes = []
    #last_memes_max = 20
    def __init__(self):
        self.last_memes = []
        self.last_memes_max = 20
    def insertMeme(self, meme):
        self.last_memes.insert(0, meme)
        if len(self.last_memes) > self.last_memes_max:
            self.last_memes.pop()
    def inQueue(self, meme) -> bool:
        return meme in self.last_memes
    def getList(self):
        return self.last_memes

recentMemes = RecentMemeQueue()

# list of all current memes' names
all_memes = []
memes_to_md5 = dict()
last_checked_all_memes = datetime.strptime("2000-01-01 01:01:01", "%Y-%m-%d %H:%M:%S")

S3_URL = ""
S3_UN  = ""
S3_PW  = ""
S3_TLS = True
S3_BUCKET = ""

client = ""

#returns the S3 client
def getClient():
    global client
    global S3_URL
    global S3_UN
    global S3_PW
    global S3_TLS
    global S3_BUCKET
    if client != "":
        return client

    if "S3_URL" not in os.environ:
        raise Exception("S3_URL is not set!")
    S3_URL = os.environ["S3_URL"]
    if "S3_UN" not in os.environ:
        raise Exception("S3_UN is not set!")
    S3_UN = os.environ["S3_UN"]
    if "S3_PW" not in os.environ:
        raise Exception("S3_PW is not set!")
    S3_PW = os.environ["S3_PW"]
    if "S3_BUCKET" not in os.environ:
        raise Exception("S3_BUCKET is not set!")
    S3_BUCKET = os.environ["S3_BUCKET"]

    # override defaults
    if "S3_TLS" in os.environ:
        S3_TLS = os.environ["S3_TLS"].lower() in ("yes", "true", "1", "t")

    if client == "":
        client = Minio(S3_URL,
                       access_key=S3_UN,
                       secret_key=S3_PW,
                       secure=S3_TLS)
        found = client.bucket_exists(S3_BUCKET)
        if not found:
            client.make_bucket(S3_BUCKET)
        else:
            print(f"Found bucket {S3_BUCKET}")

    return client

###############################################################################
# Helper methods
#  Methods to break up the calldict methods

def getCurrentMemeList(force=False):
    global last_checked_all_memes
    global all_memes
    global memes_to_md5
    now = datetime.now()
    # if no update in the past 5 mins
    if (now - last_checked_all_memes).seconds > 300 or force:
        print("Enough time has elapsed, refreshing meme cache...")
        last_checked_all_memes = now

        all_memes.clear()
        memes_to_md5.clear()

        myClient = getClient()
        for obj in myClient.list_objects(S3_BUCKET):
            if not obj.is_dir:
                #print(f'{obj.object_name}')
                all_memes.append(obj.object_name)
                memes_to_md5[obj.etag] = obj.object_name
        print(f"Got {len(all_memes)} memes")
    return all_memes

# given a file, we will return:
#  A discord.File if its smaller than 8MB
#  A catbox.moe link if its larger than 8mb
def getDiscordReadyObjectFromS3(file_name):
    size = 0
    #get the file
    client = getClient()
    stream = ""
    try:
        obj = client.get_object(bucket_name=S3_BUCKET,
                           object_name=file_name)
        stream = obj.read()
        size = len(stream)
    except:
        return "No such file " + file_name + " !"

    # check the object size
    if size >= 8000000: # to big, use catbox
        print(f"{file_name} was to big, uploading to catbox...")
        catbox_uploader = Uploader(token='')
        extension = file_name.split('.')[-1]
        if len(extension) == 0:
            extension = None
        url = catbox_uploader.upload(file_raw=stream, file_type=extension)
        return url["file"]
    else: #small enough, use discord
        print(f"{file_name} was small enough to send using discord")
        return discord.File(fp=io.BytesIO(stream), filename=file_name)

# get a meme whos name closest matches the provided query
def getCloseMemeToQuery(query, returnAllClose = False):
    global all_memes
    all_memes = getCurrentMemeList()
    top_meme = ''
    top_memes = []
    top_score = 0

    for meme in all_memes:
        current_score = fuzz.partial_ratio(query, meme)
        if current_score > top_score:
            top_meme = meme
            top_score = current_score
        if current_score == 100:
            top_memes.append(meme)
            top_meme = meme

    print("Top memes we found:")
    if len(top_memes) > 0:
        print(top_memes)
    if top_meme != '':
        print(top_meme)

    # if there was one or more perfect matches, get them
    if top_score == 100:
        top_meme = random.choice(top_memes)
        if not all(elem in recentMemes.getList() for elem in top_memes):
            while top_meme in recentMemes.getList():
                top_meme = random.choice(top_memes)
            recentMemes.insertMeme(top_meme)

    if returnAllClose and len(top_memes) > 1:
        print("Returning all found memes...")
        for meme in top_memes:
            recentMemes.insertMeme(meme)
        return top_memes, 1 if len(top_memes) <= 1 else len(top_memes)
    else:
        print(f"Returning {top_meme}...")
        recentMemes.insertMeme(top_meme)
        return top_meme, 1 if len(top_memes) <= 1 else len(top_memes)

# get the name of the user who uploaded the file
# their tag is always under the 'uploader' tag
def getMemeUploader(file_name):
    client = getClient()
    tags = client.get_object_tags(S3_BUCKET, file_name)
    if tags != None:
        uploader = tags.get("uploader")
        if uploader != None:
            return uploader
    return "No-one (probably tyler)"

# gets a random meme
# this will ensure that the meme is not in the last_memes list
def getRandomMeme():
    all_memes = getCurrentMemeList()
    choice = random.choice(all_memes)
    while recentMemes.inQueue(choice):
        choice = random.choice(all_memes)
    recentMemes.insertMeme(choice)
    return choice

# Build an underscored name from a string
def nameBuilder(command):
    uploaded_file_name = ""
    for word in command:
        uploaded_file_name += word + "_"

    # remove the last _ from the name
    uploaded_file_name = uploaded_file_name.rstrip(uploaded_file_name[-1])

    # append the file extension
    return uploaded_file_name



###############################################################################
# Calldict methods

# parse for a meme that is simliar to a given query
# There are two possibilities for parseMeme, as there is a somewhat
# complex syntax
# !meme [QUERY_STRING]
# With QUERY_STRING being optional
# due to this, we need to cover both cases
# NOTE QUERY_STRING can also be an exact filename, in which case we should
# return that instead
async def parseMeme(command, message, client):
    if len(command) > 1: # if there is a query
        query_string = ' '.join(command[1:]) #get QUERY_STRING

        #first, check if the exact meme exists
        all_memes = getCurrentMemeList()
        if query_string in all_memes: # found the exact meme, return it
            return getDiscordReadyObjectFromS3(query_string)

        #if we get here then we didnt find the meme, lets find the closest meme
        close_meme, total_close_memes = getCloseMemeToQuery(query_string)
        await message.channel.send(f"Found a close meme (from {total_close_memes} choices)")
        return getDiscordReadyObjectFromS3(close_meme)

    else: #there is no query, get a random meme
        print(f'Getting a random meme...')
        meme = getRandomMeme()
        print(f'Got a random meme ({meme})')
        return getDiscordReadyObjectFromS3(meme)

# get all ememes that are similar to the given query
async def allMemes(command, message, client):
    if len(command) > 1:
        query_string = ' '.join(command[1:])
        close_memes, total_close_memes = getCloseMemeToQuery(query_string,
                                                            returnAllClose=True)
        if total_close_memes == 0:
            return "I found nothing similar to " + query_string
        elif total_close_memes == 1:
            return getDiscordReadyObjectFromS3(close_memes)
        else:
            for meme in close_memes:
                ready_meme = getDiscordReadyObjectFromS3(meme)
                if isinstance(ready_meme, str):
                    await message.channel.send(ready_meme, reference=message)
                else:
                    await message.channel.send(file=ready_meme, reference=message)
            return "Enjoy your memes :)"
    else:
        return "I need a query!"

# get total number of memes stored
async def memeCount(command, message, client):
    all_memes = getCurrentMemeList()
    return "Currently we have " + str(len(all_memes)) + " memes"

# get a set of random memes
async def memeDump(command, message, client):
    await message.channel.send("Getting your memes...")
    count = 5
    if len(command) > 1:
        try:
            count = int(command[1])
        except:
            print("Tried to parse a number, got (" + command[1] + ")")
            count = 5

    for i in range(0,count):
        meme = getRandomMeme()
        ready_meme = getDiscordReadyObjectFromS3(meme)
        if isinstance(ready_meme, str):
            await message.channel.send(ready_meme, reference=message)
        else:
            await message.channel.send(file=ready_meme, reference=message)
    return "Enjoy your memes :)"

# get an uploaded meme and save it
async def uploadMeme(command, message, discordClient):
    global memes_to_md5

    uploader_tags = Tags.new_object_tags()
    uploader_tags["uploader"] = message.author.name

    all_memes = getCurrentMemeList()
    client = getClient()
    file_names = []

    if len(message.attachments) <= 0:
        if message.reference is not None: # this is a reply upload
            m = await message.channel.fetch_message(message.reference.message_id)
            result = await uploadMeme(command, m, client)
            return result
        else:
            return "You didn't attach anything! Please attach a file to upload"

    #TODO left off on adding logic to


    for file in message.attachments:
        # if we want to override the name
        uploaded_file_name = file.filename
        if len(command) > 1:
            uploaded_file_name = nameBuilder(command[1:])
            # add the extension
            name, ext = os.path.splitext(file.filename)
            uploaded_file_name += ext

        print("File name is (" + uploaded_file_name + ")")

        # no files of the same name
        if file.filename in all_memes:
            return "File with name '" + file.filename + "' already exists!"

        await file.save("/tmp/" + file.filename)

        result = client.fput_object(bucket_name=S3_BUCKET,
                           object_name=uploaded_file_name,
                           file_path="/tmp/" + file.filename,
                           tags=uploader_tags,
                           content_type=None)

        if result.etag in memes_to_md5:
            print("Already have that meme, etag is (" + result.etag + ")")

            client.remove_object(bucket_name=S3_BUCKET,
                    object_name=file.filename)

            return "That meme is already in the cache (it's called " + memes_to_md5[result.etag] + ")"

        os.remove("/tmp/" + file.filename)

    getCurrentMemeList(force=True)

    return "Thanks got your memes!"

async def memeBlame(command, message, discordClient):
    query_string = ""
    if len(command) <= 1:
        if message.reference is not None: # this is a reply blame
            m = await message.channel.fetch_message(message.reference.message_id)
            if m.attachments is not None and len(m.attachments) > 0:
                query_string = m.attachments[0].filename
            else: return "No file attached to replied message!"
        else: return "I need a file name or query!"
    else: query_string = ' '.join(command[1:]) #get QUERY_STRING

    return_string = ""

    all_memes = getCurrentMemeList()
    if query_string in all_memes:
        author = getMemeUploader(query_string)
        return_string = author + " uploaded " + query_string
    else:
        close_meme, total_close_memes = getCloseMemeToQuery(query_string)
        author = getMemeUploader(close_meme)
        return_string  = "I think you meant \"" + close_meme + "\", which was "
        return_string += "uploaded by " + author

    return return_string
